// pose.cpp
// DEFINOVANIE POZY OBJEKTOV, INTERPOLACIA HODNOT

#include "pose.h"

#include "assimp/matrix3x3.inl"
#include "glew.h"

ObjectPose CreatePose(float pos_x, float pos_y, float pos_z, float euler_x, float euler_y, float euler_z, float scale_x, float scale_y, float scale_z)
{
    // vytvor kvaternion na zaklade Eulerovych uhlov
    aiQuaternion rotation(3.141f * euler_y / 180, 3.141f * euler_z / 180, 3.141f * euler_x / 180);

    // prekopiruj data do objektu
    ObjectPose result;
    result.pos_x = pos_x;
    result.pos_y = pos_y;
    result.pos_z = pos_z;
    result.rotation = rotation;
    result.scale_x = scale_x;
    result.scale_y = scale_y;
    result.scale_z = scale_z;

    return result;
}


float IntepolateValuesLinear(std::vector<float> knots, std::vector<float> values, float time, bool closed)
{
    if (knots.size() < 1 || values.size() < 1 || knots.size() != values.size())
        return 0.0f;

    // najdi interval, v ktorom sa nachadza dany cas
    if (time >= (knots[knots.size() - 1]))
        return values[values.size() - 1];
    if (time <= knots[0])
        return values[0];
    int span = -1;
    while (time > knots[span + 1])
        span++;

    // mame span interval, znormalizuj cas v ramci spanu
    if (knots[span] == knots[span + 1])
        return values[span];
    float u = (time - knots[span]) / (knots[span + 1] - knots[span]);

    // interpoluj hodnoty
    return ((1 - u) * values[span] + u * values[span + 1]);
}

float IntepolateValuesCubic(std::vector<float> knots, std::vector<float> values, float time, bool closed)
{
    if (knots.size() < 1 || values.size() < 1 || knots.size() != values.size())
        return 0.0f;

    // najdi interval, v ktorom sa nachadza dany cas
    if (time >= (knots[knots.size() - 1]))
        return values[values.size() - 1];
    if (time <= knots[0])
        return values[0];
    int span = -1;
    while (time > knots[span + 1])
        span++;

    if (knots[span] == knots[span + 1])
        return values[span];

    // pre koncove hodnoty span intervalu, vypocitaj dotykove vektory
    // pouzijeme kardinalny splajn s parametrom tension
    float tension = 0;
    float prev_value = values[span];
    float prev_time = knots[span];
    if (span > 0)
    {
        prev_value = values[span - 1];
        prev_time = knots[span - 1];
    }
    float next_value = values[span + 1];
    float next_time = knots[span + 1];
    if (span < int(values.size() - 2))
    {
        next_value = values[span + 2];
        next_time = knots[span + 2];
    }
    float begin_tangent = (knots[span + 1] - prev_time) == 0 ? 0 : (1 - tension) * (values[span + 1] - prev_value) / (knots[span + 1] - prev_time);
    float end_tangent = (next_time - knots[span]) == 0 ? 0 : (1 - tension) * (next_value - values[span]) / (next_time - knots[span]);

    // dalsia moznost pre dotykove vektory na zaciatku a konci
    // budu nulove :-)
    if (span == 0)
        begin_tangent = 0;
    if (span >= int(values.size() - 2))
        end_tangent = 0;

    // mame span interval, znormalizuj cas v ramci spanu
    float u = (time - knots[span]) / (knots[span + 1] - knots[span]);

    // vypocitaj interpolovanu hodnotu na zaklade hermitovych polynomov
    float result = (2*u*u*u - 3*u*u + 1)*values[span] +
                   (u*u*u - 2*u*u + u)*(knots[span + 1] - knots[span])*begin_tangent + 
                   (-2*u*u*u + 3*u*u)*values[span + 1] + 
                   (u*u*u - u*u)*(knots[span + 1] - knots[span])*end_tangent;
    return result;
}

aiQuaternion IntepolateQuaternionsLinear(std::vector<float> knots, std::vector<aiQuaternion> quaternions, float time, bool closed)
{
    if (knots.size() < 1 || quaternions.size() < 1 || knots.size() != quaternions.size())
        return aiQuaternion(0, 0, 0);

    // najdi interval, v ktorom sa nachadza dany cas
    if (time >= (knots[knots.size() - 1]))
        return quaternions[quaternions.size() - 1];
    if (time <= knots[0])
        return quaternions[0];
    int span = -1;
    while (time > knots[span + 1])
        span++;

    // mame span interval, znormalizuj cas v ramci spanu
    if (knots[span] == knots[span + 1])
        return quaternions[span];
    float u = (time - knots[span]) / (knots[span + 1] - knots[span]);

    // interpoluj hodnoty dvoch kvaternionov
    // pomocou slerp - linearne interpolacia na sfere 
    aiQuaternion result;
    aiQuaternion::Interpolate(result, quaternions[span], quaternions[span+ 1], u);
    return result;
}

aiQuaternion DoubleArc(aiQuaternion Q1, aiQuaternion Q2)
{
    // vypocitaj inverzny kvaternion k Q1
    /*aiQuaternion Q1inverse = Q1;
    float norm = sqrt(Q1.x*Q1.x + Q1.y*Q1.y + Q1.z*Q1.z + Q1.w*Q1.w);
    norm = 1 / (norm*norm);
    Q1inverse.Conjugate();
    Q1inverse.x = norm * Q1inverse.x;
    Q1inverse.y = norm * Q1inverse.y;
    Q1inverse.z = norm * Q1inverse.z;
    Q1inverse.w = norm * Q1inverse.w;

    //vypocitaj zdvojenie obluka
    aiQuaternion Q = (Q2 * Q1inverse * Q2);
    Q.Normalize();
    return Q;*/

    // druhy sposob vypoctu zdvojenia
    float dot = Q1.x*Q2.x + Q1.y*Q2.y + Q1.z*Q2.z + Q1.w*Q2.w;
    aiQuaternion Q;
    Q.x = 2 * dot * Q2.x - Q1.x;
    Q.y = 2 * dot * Q2.y - Q1.y;
    Q.z = 2 * dot * Q2.z - Q1.z;
    Q.w = 2 * dot * Q2.w - Q1.w;
    Q.Normalize();
    return Q;
}

aiQuaternion BisectArc(aiQuaternion Q1, aiQuaternion Q2)
{
    aiQuaternion result;
    result.x = Q1.x + Q2.x;
    result.y = Q1.y + Q2.y;
    result.z = Q1.z + Q2.z;
    result.w = Q1.w + Q2.w;
    result.Normalize();

    return result;
}

aiQuaternion IntepolateQuaternionsCubic(std::vector<float> knots, std::vector<aiQuaternion> quaternions, float time, bool closed)
{
    if (knots.size() < 1 || quaternions.size() < 1 || knots.size() != quaternions.size())
        return aiQuaternion(0, 0, 0);

    // najdi interval, v ktorom sa nachadza dany cas
    if (time >= (knots[knots.size() - 1]))
        return quaternions[quaternions.size() - 1];
    if (time <= knots[0])
        return quaternions[0];
    int span = -1;
    while (time > knots[span + 1])
        span++;

    // mame span interval, znormalizuj cas v ramci spanu
    if (knots[span] == knots[span + 1])
        return quaternions[span];
    float u = (time - knots[span]) / (knots[span + 1] - knots[span]);

    // daj 4 za sebou iduce kvaterniony, pomocou ktorych ideme interpolovat medzi Qj a Qjplus1
    aiQuaternion Qjminus1 = quaternions[span];
    if (span > 0)
        Qjminus1 = quaternions[span - 1];
    aiQuaternion Qj = quaternions[span];
    aiQuaternion Qjplus1 = quaternions[span+1];
    aiQuaternion Qjplus2 = quaternions[span+1];
    if (span < int(quaternions.size() - 2))
        Qjplus2 = quaternions[span+2];

    // vypocitaj dalsie 2 pomocne bezierove vrcholy
    aiQuaternion Lj = DoubleArc(Qjminus1, Qj);
    aiQuaternion Ljplus1 = DoubleArc(Qj, Qjplus1);
    aiQuaternion Aj = BisectArc(Lj, Qjplus1);
    aiQuaternion Ajplus1 = BisectArc(Ljplus1, Qjplus2);
    aiQuaternion Bjplus1 = DoubleArc(Ajplus1, Qjplus1);

    // posun Bezierove vrcholy Aj a Bjplus1 blizsie k riadiacim vrcholom Qj a Qjplus1
    // teneto parameter definuje kvalitu interpolacnej krivky
    float tension = 0.6666f;
    Aj.x = tension * Qj.x + (1 - tension) * Aj.x;
    Aj.y = tension * Qj.y + (1 - tension) * Aj.y;
    Aj.z = tension * Qj.z + (1 - tension) * Aj.z;
    Aj.w = tension * Qj.w + (1 - tension) * Aj.w;
    Aj.Normalize();
    Bjplus1.x = tension * Qjplus1.x + (1 - tension) * Bjplus1.x;
    Bjplus1.y = tension * Qjplus1.y + (1 - tension) * Bjplus1.y;
    Bjplus1.z = tension * Qjplus1.z + (1 - tension) * Bjplus1.z;
    Bjplus1.w = tension * Qjplus1.w + (1 - tension) * Bjplus1.w;
    Bjplus1.Normalize();

    // mame bezierovy vrcholy na 4D sfere, pouzi deCasteljauov algoritmus a slerp
    aiQuaternion P1, P2, P3, P4, P5, Q;
    aiQuaternion::Interpolate(P1, Qj, Aj, u);
    aiQuaternion::Interpolate(P2, Aj, Bjplus1, u);
    aiQuaternion::Interpolate(P3, Bjplus1, Qjplus1, u);
    aiQuaternion::Interpolate(P4, P1, P2, u);
    aiQuaternion::Interpolate(P5, P2, P3, u);
    aiQuaternion::Interpolate(Q, P4, P5, u);

    return Q;
}

ObjectPose IntepolatePoseLinear(std::vector<float> knots, std::vector<ObjectPose> poses, float time, bool closed)
{
    ObjectPose result;

    // nacitaj klucove hodnoty do vektora zvlast pre kazdu suradnicu
    std::vector<float> vector_x, vector_y, vector_z;
    std::vector<float> vector_scale_x, vector_scale_y, vector_scale_z;
    std::vector<aiQuaternion> vector_rot;
    for (int i = 0; i < (int)poses.size(); i++)
    {
        vector_x.push_back(poses[i].pos_x);
        vector_y.push_back(poses[i].pos_y);
        vector_z.push_back(poses[i].pos_z);
        vector_scale_x.push_back(poses[i].scale_x);
        vector_scale_y.push_back(poses[i].scale_y);
        vector_scale_z.push_back(poses[i].scale_z);
        vector_rot.push_back(poses[i].rotation);
    }

    // interpoluj kazdu suradnicu separatne
    result.pos_x = IntepolateValuesLinear(knots, vector_x, time, closed);
    result.pos_y = IntepolateValuesLinear(knots, vector_y, time, closed);
    result.pos_z = IntepolateValuesLinear(knots, vector_z, time, closed);
    result.scale_x = IntepolateValuesLinear(knots, vector_scale_x, time, closed);
    result.scale_y = IntepolateValuesLinear(knots, vector_scale_y, time, closed);
    result.scale_z = IntepolateValuesLinear(knots, vector_scale_z, time, closed);
    result.rotation = IntepolateQuaternionsLinear(knots, vector_rot, time, closed);

    return result;
}

ObjectPose IntepolatePoseCubic(std::vector<float> knots, std::vector<ObjectPose> poses, float time, bool closed)
{
    ObjectPose result;

    // nacitaj klucove hodnoty do vektora zvlast pre kazdu suradnicu
    std::vector<float> vector_x, vector_y, vector_z;
    std::vector<float> vector_scale_x, vector_scale_y, vector_scale_z;
    std::vector<aiQuaternion> vector_rot;
    for (int i = 0; i < (int)poses.size(); i++)
    {
        vector_x.push_back(poses[i].pos_x);
        vector_y.push_back(poses[i].pos_y);
        vector_z.push_back(poses[i].pos_z);
        vector_scale_x.push_back(poses[i].scale_x);
        vector_scale_y.push_back(poses[i].scale_y);
        vector_scale_z.push_back(poses[i].scale_z);
        vector_rot.push_back(poses[i].rotation);
    }

    // interpoluj kazdy realny parameter separatne
    result.pos_x = IntepolateValuesCubic(knots, vector_x, time, closed);
    result.pos_y = IntepolateValuesCubic(knots, vector_y, time, closed);
    result.pos_z = IntepolateValuesCubic(knots, vector_z, time, closed);
    result.scale_x = IntepolateValuesCubic(knots, vector_scale_x, time, closed);
    result.scale_y = IntepolateValuesCubic(knots, vector_scale_y, time, closed);
    result.scale_z = IntepolateValuesCubic(knots, vector_scale_z, time, closed);
    result.rotation = IntepolateQuaternionsCubic(knots, vector_rot, time, closed);

    return result;
}

void CrossProduct(float v1_x, float v1_y, float v1_z, float v2_x, float v2_y, float v2_z, float& out_x, float& out_y, float& out_z) 
{
    out_x = v1_y*v2_z - v1_z*v2_y;
    out_y = v1_z*v2_x - v1_x*v2_z;
    out_z = v1_x*v2_y - v1_y*v2_x;
}

ObjectPose IntepolatePoseCubicDerivedRot(std::vector<float> knots, std::vector<ObjectPose> poses, aiQuaternion base, float time, bool closed)
{
    ObjectPose result;
    if (poses.size() < 1 || knots.size() < 1)
        return result;

    // nacitaj klucove hodnoty do vektora zvlast pre kazdu suradnicu
    std::vector<float> vector_x, vector_y, vector_z;
    std::vector<float> vector_scale_x, vector_scale_y, vector_scale_z;
    std::vector<aiQuaternion> vector_rot;
    for (int i = 0; i < (int)poses.size(); i++)
    {
        vector_x.push_back(poses[i].pos_x);
        vector_y.push_back(poses[i].pos_y);
        vector_z.push_back(poses[i].pos_z);
        vector_scale_x.push_back(poses[i].scale_x);
        vector_scale_y.push_back(poses[i].scale_y);
        vector_scale_z.push_back(poses[i].scale_z);
        vector_rot.push_back(poses[i].rotation);
    }

    // interpoluj kazdy realny parameter separatne
    result.pos_x = IntepolateValuesCubic(knots, vector_x, time, closed);
    result.pos_y = IntepolateValuesCubic(knots, vector_y, time, closed);
    result.pos_z = IntepolateValuesCubic(knots, vector_z, time, closed);
    result.scale_x = IntepolateValuesCubic(knots, vector_scale_x, time, closed);
    result.scale_y = IntepolateValuesCubic(knots, vector_scale_y, time, closed);
    result.scale_z = IntepolateValuesCubic(knots, vector_scale_z, time, closed);

    // rotacia je urcena derivaciou interpolacnej krivky
    // zatial sa derivacia vypocita numericky, nie z analytickeho vyjadrenia
    float prev_x = IntepolateValuesCubic(knots, vector_x, time - 0.001f, closed);
    float prev_y = IntepolateValuesCubic(knots, vector_y, time - 0.001f, closed);
    float prev_z = IntepolateValuesCubic(knots, vector_z, time - 0.001f, closed);
    float next_x = IntepolateValuesCubic(knots, vector_x, time + 0.001f, closed);
    float next_y = IntepolateValuesCubic(knots, vector_y, time + 0.001f, closed);
    float next_z = IntepolateValuesCubic(knots, vector_z, time + 0.001f, closed);
    float dir_x = next_x - prev_x;
    float dir_y = next_y - prev_y;
    float dir_z = next_z - prev_z;
    float length = sqrt(dir_x * dir_x + dir_y * dir_y + dir_z * dir_z);
    if (length != 0)
    {
        dir_x = dir_x / length;
        dir_y = dir_y / length;
        dir_z = dir_z / length;

        float up_x = 0;
        float up_y = 0;
        float up_z = 1;

        // vektorovy sucin pre direction a up vektory dava right vektor
        float right_x, right_y, right_z;
        CrossProduct(dir_x, dir_y, dir_z, up_x, up_y, up_z, right_x, right_y, right_z);
        length = sqrt(right_x * right_x + right_y * right_y + right_z * right_z);
        right_x = right_x / length;
        right_y = right_y / length;
        right_z = right_z / length;

        // vektorovy sucin pre direction a up vektory dava right vektor
        float new_up_x, new_up_y, new_up_z;
        CrossProduct(right_x, right_y, right_z, dir_x, dir_y, dir_z, new_up_x, new_up_y, new_up_z);
        
        float d1 = right_x * dir_x + right_y * dir_y + right_z * dir_z;
        length = sqrt(new_up_x * new_up_x + new_up_y * new_up_y + new_up_z * new_up_z);

        // tieto 3 vektory tvoria stlpce rotacnej matice
        aiMatrix3x3 rotation_matrix(right_x, dir_x, new_up_x,
                                    right_y, dir_y, new_up_y,
                                    right_z, dir_z, new_up_z);
        // vyrob kvaternion z rotacnej matice
        result.rotation = aiQuaternion(rotation_matrix) * base;
        result.rotation.Normalize();
    }
    else
        result.rotation = IntepolateQuaternionsCubic(knots, vector_rot, time, closed);

    return result;
}

aiMatrix4x4 GetMatrix(ObjectPose pose)
{
    // vyrob maticu posunutia
    aiVector3D translation(pose.pos_x, pose.pos_y, pose.pos_z);
    aiMatrix4x4 translationMatrix;
    aiMatrix4x4::Translation(translation, translationMatrix);

    // vyrob maticu skalovania
    aiVector3D scale(pose.scale_x, pose.scale_y, pose.scale_z);
    aiMatrix4x4 scaleMatrix;
    aiMatrix4x4::Scaling(scale, scaleMatrix);

    // vyrob maticu rotacie
    aiMatrix4x4 rotationMatrix(pose.rotation.GetMatrix());

    return (translationMatrix * rotationMatrix * scaleMatrix);
    //return (translationMatrix * scaleMatrix * rotationMatrix);
}

void MultiplyMatrixOpenGL(aiMatrix4x4 matrix)
{
    float opengl_matrix[16];

    opengl_matrix[0] = matrix.a1;
    opengl_matrix[1] = matrix.b1;
    opengl_matrix[2] = matrix.c1;
    opengl_matrix[3] = matrix.d1;
    opengl_matrix[4] = matrix.a2;
    opengl_matrix[5] = matrix.b2;
    opengl_matrix[6] = matrix.c2;
    opengl_matrix[7] = matrix.d2;
    opengl_matrix[8] = matrix.a3;
    opengl_matrix[9] = matrix.b3;
    opengl_matrix[10] = matrix.c3;
    opengl_matrix[11] = matrix.d3;
    opengl_matrix[12] = matrix.a4;
    opengl_matrix[13] = matrix.b4;
    opengl_matrix[14] = matrix.c4;
    opengl_matrix[15] = matrix.d4;

    glMultMatrixf(opengl_matrix);
}

ObjectPose CreateLookAtPose(float from_x, float from_y, float from_z, float to_x, float to_y, float to_z, float up_x, float up_y, float up_z)
{
    ObjectPose result;
    
    // skalovanie je jednotkove, netreba menit
    // pozicia je dana jasne
    result.pos_x = from_x;
    result.pos_y = from_y;
    result.pos_z = from_z;

    // vyrob vektory predstavujuce suradnice bazickych vektorov rotacie vo svetovom systeme
    float dir_x = from_x - to_x;
    float dir_y = from_y - to_y;
    float dir_z = from_z - to_z;
    float length = sqrt(dir_x * dir_x + dir_y * dir_y + dir_z * dir_z);
    if (length == 0)
        return result;
    dir_x = dir_x / length;
    dir_y = dir_y / length;
    dir_z = dir_z / length;

    // vektorovy sucin pre direction a up vektory dava right vektor
    float right_x, right_y, right_z;
    CrossProduct(up_x, up_y, up_z, dir_x, dir_y, dir_z, right_x, right_y, right_z);
    length = sqrt(right_x * right_x + right_y * right_y + right_z * right_z);
    right_x = right_x / length;
    right_y = right_y / length;
    right_z = right_z / length;

    // vektorovy sucin pre direction a up vektory dava right vektor
    float new_up_x, new_up_y, new_up_z;
    CrossProduct(dir_x, dir_y, dir_z, right_x, right_y, right_z, new_up_x, new_up_y, new_up_z);
    
    float d1 = right_x * dir_x + right_y * dir_y + right_z * dir_z;
    length = sqrt(new_up_x * new_up_x + new_up_y * new_up_y + new_up_z * new_up_z);

    // tieto 3 vektory tvoria stlpce rotacnej matice
    aiMatrix3x3 rotation_matrix(right_x, new_up_x, dir_x,
                                right_y, new_up_y, dir_y,
                                right_z, new_up_z, dir_z);
    // vyrob kvaternion z rotacnej matice
    aiQuaternion rotation(rotation_matrix);
    result.rotation = rotation;

    return result;
}